@@ -3,9 +3,10 @@
 #
 
 import re
-from Errors import error, InternalError, warning
+from Errors import warning, error, InternalError
 import Options
 import Naming
+import PyrexTypes
 from PyrexTypes import *
 import TypeSlots
 from TypeSlots import \
@@ -24,16 +25,16 @@ class Entry:
     # doc              string     Doc string
     # init             string     Initial value
     # visibility       'private' or 'public' or 'extern'
-    # is_builtin       boolean    Is a Python builtin name
+    # is_builtin       boolean    Is an entry in the Python builtins dict
     # is_cglobal       boolean    Is a C global variable
     # is_pyglobal      boolean    Is a Python module-level variable
     #                               or class attribute during
     #                               class construction
-    # is_special       boolean    Is a special class method
     # is_member        boolean    Is an assigned class member
     # is_variable      boolean    Is a variable
     # is_cfunction     boolean    Is a C function
     # is_cmethod       boolean    Is a C method of an extension type
+    # is_unbound_cmethod boolean  Is an unbound C method of an extension type
     # is_type          boolean    Is a type definition
     # is_const         boolean    Is a constant
     # is_property      boolean    Is a property of an extension type:
@@ -51,7 +52,7 @@ class Entry:
     # signature        Signature  Arg & return types for Python func
     # init_to_none     boolean    True if initial value should be None
     # as_variable      Entry      Alternative interpretation of extension
-    #                               type name as a variable
+    #                               type name or builtin C function as a variable
     # xdecref_cleanup  boolean    Use Py_XDECREF for error cleanup
     # in_cinclude      boolean    Suppress C declaration code
     # enum_values      [Entry]    For enum types, list of values
@@ -61,10 +62,15 @@ class Entry:
     #                                 type is an extension type
     # as_module        None       Module scope, if a cimported module
     # is_inherited     boolean    Is an inherited attribute of an extension type
-    # interned_cname   string     C name of interned name string
+    # #interned_cname   string     C name of interned name string
     # pystring_cname   string     C name of Python version of string literal
     # is_interned      boolean    For string const entries, value is interned
     # used             boolean
+    # is_special       boolean    Is a special method or property accessor
+    #                               of an extension type
+    # defined_in_pxd   boolean    Is defined in a .pxd file (not just declared)
+    # api              boolean    Generate C API for C class or function
+    # utility_code     string     Utility code needed when this entry is used
 
     borrowed = 0
     init = ""
@@ -72,11 +78,11 @@ class Entry:
     is_builtin = 0
     is_cglobal = 0
     is_pyglobal = 0
-    is_special = 0
     is_member = 0
     is_variable = 0
     is_cfunction = 0
     is_cmethod = 0
+    is_unbound_cmethod = 0
     is_type = 0
     is_const = 0
     is_property = 0
@@ -95,11 +101,16 @@ class Entry:
     in_cinclude = 0
     as_module = None
     is_inherited = 0
-    interned_cname = None
+    #interned_cname = None
     pystring_cname = None
     is_interned = 0
     used = 0
-    
+    is_special = 0
+    defined_in_pxd = 0
+    api = 0
+    utility_code = None
+    is_overridable = 0
+
     def __init__(self, name, cname, type, pos = None, init = None):
         self.name = name
         self.cname = cname
@@ -113,6 +124,7 @@ class Scope:
     # outer_scope       Scope or None      Enclosing scope
     # entries           {string : Entry}   Python name to entry, non-types
     # const_entries     [Entry]            Constant entries
+    # type_entries      [Entry]            Struct/union/enum/typedef/exttype entries
     # sue_entries       [Entry]            Struct/union/enum entries
     # arg_entries       [Entry]            Function argument entries
     # var_entries       [Entry]            User-defined variable entries
@@ -155,6 +167,7 @@ class Scope:
             self.scope_prefix = mangled_name
         self.entries = {}
         self.const_entries = []
+        self.type_entries = []
         self.sue_entries = []
         self.arg_entries = []
         self.var_entries = []
@@ -236,27 +249,42 @@ class Scope:
         return entry
     
     def declare_type(self, name, type, pos, 
-            cname = None, visibility = 'private'):
+            cname = None, visibility = 'private', defining = 1):
         # Add an entry for a type definition.
         if not cname:
             cname = name
         entry = self.declare(name, cname, type, pos)
         entry.visibility = visibility
         entry.is_type = 1
+        if defining:
+            self.type_entries.append(entry)
+        return entry
+    
+    def declare_typedef(self, name, base_type, pos, cname = None,
+            visibility = 'private'):
+        if not cname:
+            if self.in_cinclude or visibility == 'public':
+                cname = name
+            else:
+                cname = self.mangle(Naming.type_prefix, name)
+        type = PyrexTypes.CTypedefType(cname, base_type)
+        entry = self.declare_type(name, type, pos, cname, visibility)
+        type.qualified_name = entry.qualified_name
         return entry
         
     def declare_struct_or_union(self, name, kind, scope, 
-            typedef_flag, pos, cname = None):
+            typedef_flag, pos, cname = None, visibility = 'private'):
         # Add an entry for a struct or union definition.
         if not cname:
-            if self.in_cinclude:
+            if self.in_cinclude or visibility == 'public':
                 cname = name
             else:
                 cname = self.mangle(Naming.type_prefix, name)
         entry = self.lookup_here(name)
         if not entry:
             type = CStructOrUnionType(name, kind, scope, typedef_flag, cname)
-            entry = self.declare_type(name, type, pos, cname)
+            entry = self.declare_type(name, type, pos, cname,
+                visibility = visibility, defining = scope is not None)
             self.sue_entries.append(entry)
         else:
             if not (entry.is_type and entry.type.is_struct_or_union):
@@ -265,8 +293,10 @@ class Scope:
                 warning(pos, "'%s' already defined  (ignoring second definition)" % name, 0)
             else:
                 self.check_previous_typedef_flag(entry, typedef_flag, pos)
+                self.check_previous_visibility(entry, visibility, pos)
                 if scope:
                     entry.type.scope = scope
+                    self.type_entries.append(entry)
         if not scope and not entry.type.scope:
             self.check_for_illegal_incomplete_ctypedef(typedef_flag, pos)
         return entry
@@ -276,17 +306,24 @@ class Scope:
             error(pos, "'%s' previously declared using '%s'" % (
                 entry.name, ("cdef", "ctypedef")[entry.type.typedef_flag]))
     
-    def declare_enum(self, name, pos, cname, typedef_flag):
+    def check_previous_visibility(self, entry, visibility, pos):
+        if entry.visibility <> visibility:
+            error(pos, "'%s' previously declared as '%s'" % (
+                entry.name, entry.visibility))
+    
+    def declare_enum(self, name, pos, cname, typedef_flag,
+            visibility = 'private'):
         if name:
             if not cname:
-                if self.in_cinclude:
+                if self.in_cinclude or visibility == 'public':
                     cname = name
                 else:
                     cname = self.mangle(Naming.type_prefix, name)
             type = CEnumType(name, cname, typedef_flag)
         else:
-            type = c_int_type
-        entry = self.declare_type(name, type, pos, cname = cname)
+            type = PyrexTypes.c_anon_enum_type
+        entry = self.declare_type(name, type, pos, cname = cname,
+            visibility = visibility)
         entry.enum_values = []
         self.sue_entries.append(entry)
         return entry	
@@ -318,15 +355,29 @@ class Scope:
         self.pyfunc_entries.append(entry)
     
     def declare_cfunction(self, name, type, pos, 
-            cname = None, visibility = 'private', defining = 0):
+            cname = None, visibility = 'private', defining = 0, api = 0, in_pxd = 0):
         # Add an entry for a C function.
-        if not cname:
-            if visibility <> 'private':
-                cname = name
-            else:
-                cname = self.mangle(Naming.func_prefix, name)
-        entry = self.add_cfunction(name, type, pos, cname, visibility)
-        entry.func_cname = cname
+        entry = self.lookup_here(name)
+        if entry:
+            if visibility <> 'private' and visibility <> entry.visibility:
+                warning(pos, "Function '%s' previously declared as '%s'" % (name, entry.visibility), 1)
+            if not entry.type.same_as(type):
+                warning(pos, "Function signature does not match previous declaration", 1)
+                entry.type = type
+        else:
+            if not cname:
+                if api or visibility <> 'private':
+                    cname = name
+                else:
+                    cname = self.mangle(Naming.func_prefix, name)
+            entry = self.add_cfunction(name, type, pos, cname, visibility)
+            entry.func_cname = cname
+        if in_pxd and visibility <> 'extern':
+            entry.defined_in_pxd = 1
+        if api:
+            entry.api = 1
+        if not defining and not in_pxd and visibility <> 'extern':
+            error(pos, "Non-extern C function declared but not defined")
         return entry
     
     def add_cfunction(self, name, type, pos, cname, visibility):
@@ -401,7 +452,7 @@ class Scope:
         # Add an entry for an int constant.
         cname = "%s%s" % (Naming.interned_num_prefix, value)
         cname = cname.replace('-', 'neg_').replace('.','_')
-        entry = Entry("", cname, c_long_type, init = value)
+        entry = Entry("", cname, py_object_type, init = value)
         entry.used = 1
         entry.is_interned = 1
         self.const_entries.append(entry)
@@ -512,8 +563,6 @@ class BuiltinScope(Scope):
         for name, definition in self.builtin_entries.iteritems():
             cname, type = definition
             self.declare_var(name, type, None, cname)
-        self.cached_entries = []
-        self.undeclared_cached_entries = []
         
     def declare_builtin(self, name, pos):
         if not hasattr(__builtin__, name):
@@ -521,41 +570,34 @@ class BuiltinScope(Scope):
                 return self.outer_scope.declare_builtin(name, pos)
             else:
                 error(pos, "undeclared name not builtin: %s"%name)
-        entry = self.declare(name, name, py_object_type, pos)
-        if Options.cache_builtins:
-            entry.is_builtin = 1
-            entry.is_const = 1
-            entry.cname = Naming.builtin_prefix + name
-            self.cached_entries.append(entry)
-            self.undeclared_cached_entries.append(entry)
-        else:
-            entry.is_builtin = 1
+    
+    def declare_builtin_cfunction(self, name, type, cname, python_equiv = None,
+            utility_code = None):
+        # If python_equiv == "*", the Python equivalent has the same name
+        # as the entry, otherwise it has the name specified by python_equiv.
+        entry = self.declare_cfunction(name, type, None, cname)
+        entry.utility_code = utility_code
+        if python_equiv:
+            if python_equiv == "*":
+                python_equiv = name
+            var_entry = Entry(python_equiv, python_equiv, py_object_type)
+            var_entry.is_variable = 1
+            var_entry.is_builtin = 1
+            entry.as_variable = var_entry
         return entry
-        
+
     def builtin_scope(self):
         return self
         
-    # TODO: built in functions conflict with built in types of same name...
-    # TODO: perhapse these should all be declared in some universal .pxi file? 
+    # TODO: merge this into builtin_function_table when error handling in Pyrex
+    #       is fixed. Also handle pyrex types as functions. 
     
     builtin_functions = {
-      "hasattr": ["PyObject_HasAttr", c_bint_type, (py_object_type, py_object_type)],
-      "setattr": ["PyObject_SetAttr", c_int_type, (py_object_type, py_object_type, py_object_type), -1],
       "cmp":     ["PyObject_Compare", c_int_type, (py_object_type, py_object_type), None, True],
-      "repr":    ["PyObject_Repr", py_object_type, (py_object_type, ), 0],
-#      "str":     ["PyObject_Str", py_object_type, (py_object_type, ), 0],
       "unicode": ["PyObject_Unicode", py_object_type, (py_object_type, ), 0],
-      "isinstance": ["PyObject_IsInstance", c_bint_type, (py_object_type, py_object_type), -1],
-      "issubclass": ["PyObject_IsSubclass", c_bint_type, (py_object_type, py_object_type), -1],
-      "hash":    ["PyObject_Hash", c_long_type, (py_object_type, ), -1, True],
       "type":    ["PyObject_Type", py_object_type, (py_object_type, ), 0],
-      "len":     ["PyObject_Size", c_py_ssize_t_type, (py_object_type, ), -1],
-      "dir":     ["PyObject_Dir", py_object_type, (py_object_type, ), 0],
-      "iter":    ["PyObject_GetIter", py_object_type, (py_object_type, ), 0],
 
-      "abs":     ["PyNumber_Absolute", py_object_type, (py_object_type, ), 0],
-      "divmod":  ["PyNumber_Divmod", py_object_type, (py_object_type, py_object_type), 0],
-      "pow":     ["PyNumber_Power", py_object_type, (py_object_type, py_object_type, py_object_type), 0],
+#      "str":     ["PyObject_Str", py_object_type, (py_object_type, ), 0],
 #      "int":     ["PyNumber_Int", py_object_type, (py_object_type, ), 0],
 #      "long":    ["PyNumber_Long", py_object_type, (py_object_type, ), 0],
 #      "float":   ["PyNumber_Float", py_object_type, (py_object_type, ), 0],
@@ -580,7 +622,7 @@ class BuiltinScope(Scope):
         "type":   ["((PyObject*)&PyType_Type)", py_object_type],
         "slice":  ["((PyObject*)&PySlice_Type)", py_object_type],
         "file":   ["((PyObject*)&PyFile_Type)", py_object_type],
-        
+
         "None":   ["Py_None", py_object_type],
         "False":  ["Py_False", py_object_type],
         "True":   ["Py_True", py_object_type],
@@ -636,6 +678,9 @@ class ModuleScope(Scope):
         self.all_pystring_entries = []
         self.types_imported = {}
         self.pynum_entries = []
+        self.has_extern_class = 0
+        self.cached_builtins = []
+        self.undeclared_cached_builtins = []
     
     def qualifying_scope(self):
         return self.parent_module
@@ -644,8 +689,20 @@ class ModuleScope(Scope):
         return self
     
     def declare_builtin(self, name, pos):
-        entry = Scope.declare_builtin(self, name, pos)
-        entry.interned_cname = self.intern(name)
+        if not hasattr(__builtin__, name):
+            if self.outer_scope is not None:
+                return self.outer_scope.declare_builtin(name, pos)
+            else:
+                error(pos, "undeclared name not builtin: %s"%name)
+        entry = self.declare(name, name, py_object_type, pos)
+        if Options.cache_builtins:
+            entry.is_builtin = 1
+            entry.is_const = 1
+            entry.cname = Naming.builtin_prefix + name
+            self.cached_builtins.append(entry)
+            self.undeclared_cached_builtins.append(entry)
+        else:
+            entry.is_builtin = 1
         return entry
     
     def intern(self, name):
@@ -736,8 +793,8 @@ class ModuleScope(Scope):
                     "Non-cdef global variable is not a generic Python object")
             entry.is_pyglobal = 1
             entry.namespace_cname = self.module_cname
-            if Options.intern_names:
-                entry.interned_cname = self.intern(name)
+            #if Options.intern_names:
+            #	entry.interned_cname = self.intern(name)
         else:
             entry.is_cglobal = 1
             self.var_entries.append(entry)
@@ -772,10 +829,7 @@ class ModuleScope(Scope):
     
     def declare_c_class(self, name, pos, defining, implementing,
         module_name, base_type, objstruct_cname, typeobj_cname,
-        visibility, typedef_flag):
-        #
-        #print "declare_c_class:", name
-        #print "...visibility =", visibility
+        visibility, typedef_flag, api):
         #
         # Look for previous declaration as a type
         #
@@ -798,7 +852,8 @@ class ModuleScope(Scope):
             else:
                 type.module_name = self.qualified_name
             type.typeptr_cname = self.mangle(Naming.typeptr_prefix, name)
-            entry = self.declare_type(name, type, pos, visibility = visibility)
+            entry = self.declare_type(name, type, pos, visibility = visibility,
+                defining = 0)
             if objstruct_cname:
                 type.objstruct_cname = objstruct_cname
             elif not entry.in_cinclude:
@@ -818,6 +873,7 @@ class ModuleScope(Scope):
                 if base_type:
                     scope.declare_inherited_c_attributes(base_type.scope)
                 type.set_scope(scope)
+                self.type_entries.append(entry)
             else:
                 self.check_for_illegal_incomplete_ctypedef(typedef_flag, pos)
         else:
@@ -828,11 +884,15 @@ class ModuleScope(Scope):
         #
         # Fill in options, checking for compatibility with any previous declaration
         #
+        if defining:
+            entry.defined_in_pxd = 1
         if implementing:   # So that filenames in runtime exceptions refer to
             entry.pos = pos  # the .pyx file and not the .pxd file
-        if entry.visibility <> visibility:
-            error(pos, "Declaration of '%s' as '%s' conflicts with previous "
-                "declaration as '%s'" % (class_name, visibility, entry.visibility))
+        if visibility <> 'private' and entry.visibility <> visibility:
+            error(pos, "Class '%s' previously declared as '%s'"
+                % (name, entry.visibility))
+        if api:
+            entry.api = 1
         if objstruct_cname:
             if type.objstruct_cname and type.objstruct_cname <> objstruct_cname:
                 error(pos, "Object struct name differs from previous declaration")
@@ -989,6 +1049,9 @@ class StructOrUnionScope(Scope):
                 "C struct/union member cannot be declared %s" % visibility)
         return entry
 
+    def declare_cfunction(self, name, type, pos, 
+            cname = None, visibility = 'private', defining = 0, api = 0, in_pxd = 0):
+        self.declare_var(name, type, pos, cname, visibility)
 
 class ClassScope(Scope):
     #  Abstract base class for namespace of
@@ -1023,8 +1086,8 @@ class PyClassScope(ClassScope):
             cname, visibility, is_cdef)
         entry.is_pyglobal = 1
         entry.namespace_cname = self.class_obj_cname
-        if Options.intern_names:
-            entry.interned_cname = self.intern(name)
+        #if Options.intern_names:
+        #	entry.interned_cname = self.intern(name)
         return entry
 
     def allocate_temp(self, type):
@@ -1131,6 +1194,10 @@ class CClassScope(ClassScope):
         # Add an entry for a method.
         if name in ('__eq__', '__ne__', '__lt__', '__gt__', '__le__', '__ge__'):
             error(pos, "Special method %s must be implemented via __richcmp__" % name)
+        if name == "__new__":
+            warning(pos, "__new__ method of extension type will change semantics "
+                "in a future version of Pyrex and Cython. Use __cinit__ instead.")
+            name = "__cinit__"
         entry = self.declare_var(name, py_object_type, pos)
         special_sig = get_special_method_signature(name)
         if special_sig:
@@ -1144,9 +1211,14 @@ class CClassScope(ClassScope):
 
         self.pyfunc_entries.append(entry)
         return entry
-            
+    
+    def lookup_here(self, name):
+        if name == "__new__":
+            name = "__cinit__"
+        return ClassScope.lookup_here(self, name)
+    
     def declare_cfunction(self, name, type, pos,
-            cname = None, visibility = 'private', defining = 0):
+            cname = None, visibility = 'private', defining = 0, api = 0, in_pxd = 0):
         if get_special_method_signature(name):
             error(pos, "Special methods must be declared with 'def', not 'cdef'")
         args = type.args
@@ -1161,6 +1233,7 @@ class CClassScope(ClassScope):
             else:
                 if defining and entry.func_cname:
                     error(pos, "'%s' already defined" % name)
+                #print "CClassScope.declare_cfunction: checking signature" ###
                 if type.same_c_signature_as(entry.type, as_cmethod = 1):
                     pass
 #                if type.narrower_c_signature_than(entry.type, as_cmethod = 1):
@@ -1243,8 +1316,8 @@ class PropertyScope(Scope):
         signature = get_property_accessor_signature(name)
         if signature:
             entry = self.declare(name, name, py_object_type, pos)
-            entry.signature = signature
             entry.is_special = 1
+            entry.signature = signature
             return entry
         else:
             error(pos, "Only __get__, __set__ and __del__ methods allowed "
