@@ -16,6 +16,9 @@ setup(
 #    Extension("inherit_bug", ["inherit_bug.pyx"]),
     Extension("override", ["override.pyx"]),
     Extension("cond", ["cond.pyx"]),
+    Extension("submodule.test",       ["submodule/test.pyx"]),
+    Extension("errors",       ["errors.pyx"]),
+    Extension("cpdef",       ["cpdef.pyx"]),
     ],
   cmdclass = {'build_ext': build_ext},
 #  include_dirs = "/System/Library/Frameworks/Python.framework/Versions/2.3/include/python2.3/"
